import unittest
import ast

from ctree.transformations import PyBasicConversions
from ctree.c.nodes import *


class TestAssigns(unittest.TestCase):
    def setUp(self):
        self.foo, self.bar = SymbolRef("foo"), SymbolRef("bar")

    def test_simple_assign(self):
        node = Assign(self.foo, self.bar)
        self.assertEqual(str(node), "foo = bar")


    def test_multiple_assign_simple(self):
        node = ast.Assign([ast.Tuple(elts = (ast.Name(id = "x", ctx = None), ast.Name(id = "y", ctx = None)))], ast.Tuple(elts = (ast.Name(id = "x", ctx = None), ast.Name(id = "y", ctx = None))))
        transformed_node = PyBasicConversions().visit(node)

        self.assertEqual(str(transformed_node), "\n____temp__x = x;\n____temp__y = y;\nx = ____temp__x;\ny = ____temp__y;\n")

    def test_multiple_assign_constant(self):
        node = ast.Assign([ast.Tuple(elts = (ast.Name(id = "x", ctx = None), ast.Name(id = "y", ctx = None)))], ast.Tuple(elts = (Constant(1), Constant(2))))
        transformed_node = PyBasicConversions().visit(node)

        self.assertEqual(str(transformed_node), "\nx = 1;\ny = 2;\n")

    def test_multiple_assign_dependent(self):
        node = ast.Assign([ast.Tuple(elts = (ast.Name(id = "x", ctx = None), ast.Name(id = "y", ctx = None)))], ast.Tuple(elts = (ast.Name(id = "y", ctx = None), ast.Name(id = "x", ctx = None))))
        transformed_node = PyBasicConversions().visit(node)

        self.assertEqual(str(transformed_node), "\n____temp__x = x;\n____temp__y = y;\ny = ____temp__y;\nx = ____temp__x;\n")

    def test_multiple_assign_dependent(self):
        node = ast.Assign([ast.Tuple(elts = (ast.Name(id = "x", ctx = None), ast.Name(id = "y", ctx = None)))], ast.Tuple(elts = (FunctionCall(func = 'square', args = [Constant(5), Constant(5)]), FunctionCall(func = 'square', args = [Constant(5), Constant(5)]))))
        transformed_node = PyBasicConversions().visit(node)

        self.assertEqual(str(transformed_node), "\n____temp__x = square(5, 5);\n____temp__y = square(5, 5);\nx = ____temp__x;\ny = ____temp__y;\n")



from ctypes import *

from util import CtreeTest
from ctree.c.nodes import *


class TestFuncDecls(CtreeTest):
    def test_voidvoid(self):
        node = FunctionDecl(c_void_p(), SymbolRef("foo"))
        self._check_code(node, "void* foo()")

    def test_intvoid(self):
        node = FunctionDecl(c_int(), SymbolRef("foo"))
        self._check_code(node, "int foo()")

    def test_voidint(self):
        params = [SymbolRef("a", c_int())]
        node = FunctionDecl(c_void_p(), SymbolRef("foo"), params)
        self._check_code(node, "void* foo(int a)")

    def test_intint(self):
        params = [SymbolRef("b", c_int())]
        node = FunctionDecl(c_int(), SymbolRef("foo"), params)
        self._check_code(node, "int foo(int b)")

    def test_voidintint(self):
        params = [SymbolRef("c", c_int()), SymbolRef("d", c_int())]
        node = FunctionDecl(c_void_p(), SymbolRef("foo"), params)
        self._check_code(node, "void* foo(int c, int d)")

    def test_voidintint_names(self):
        params = [SymbolRef("bar", c_int()), SymbolRef('baz', c_int())]
        node = FunctionDecl(c_void_p(), SymbolRef("foo"), params)
        self._check_code(node, "void* foo(int bar, int baz)")

    def test_withdefn(self):
        body = [Add(SymbolRef('foo'), SymbolRef('bar'))]
        node = FunctionDecl(c_void_p(), SymbolRef("fn"), defn=body)
        self._check_code(node, """\
        void* fn() {
            foo + bar;
        }""")

    def test_set_kernel(self):
        params = [SymbolRef("bar", c_int()), SymbolRef('baz', c_int())]
        node = FunctionDecl(c_void_p(), SymbolRef("foo"), params)
        node.set_kernel();
        self._check_code(node, "__kernel void* foo(int bar, int baz)")



"""
AST nodes for C constructs.
"""

import os
import types
import subprocess

import logging

log = logging.getLogger(__name__)

from ctypes import CFUNCTYPE
from ctree.nodes import CtreeNode, File
import ctree
from ctree.util import singleton, highlight, truncate
from ctree.types import get_ctype, get_common_ctype
import hashlib


class CNode(CtreeNode):
    """Base class for all C nodes in ctree."""

    def codegen(self, indent=0):
        from ctree.c.codegen import CCodeGen
        from ctree.transformations import DeclarationFiller

        return CCodeGen(indent).visit(self)

    def label(self):
        from ctree.c.dotgen import CDotGenLabeller

        return CDotGenLabeller().visit(self)


class CFile(CNode, File):
    """Represents a .c file."""
    _ext = "c"

    def __init__(self, name="generated", body=None, config_target='c', path = None):
        CNode.__init__(self)
        File.__init__(self, name, body, path)
        self.config_target = config_target

    def get_bc_filename(self):
        return "%s.bc" % self.name

    def get_so_filename(self):
        return "{}.so".format(self.name)

    def _compile(self, program_text):
        # print(repr(self.path), repr(self.get_filename()))
        c_src_file = os.path.join(self.path, self.get_filename())
        so_file = os.path.join(self.path, self.get_so_filename())
        program_hash = hashlib.sha512(program_text.strip().encode()).hexdigest()
        so_file_exists = os.path.exists(so_file)
        old_hash = self.program_hash
        hash_match = old_hash == program_hash
        log.debug("Old hash: %s \n New hash: %s", old_hash, program_hash)
        recreate_c_src = program_text and program_text != self.empty and not hash_match
        recreate_so = recreate_c_src or not so_file_exists

        log.debug("RECREATE_C_SRC: %s \t RECREATE_so: %s \t HASH_MATCH: %s",
                 recreate_c_src, recreate_so, hash_match)

        if not program_text:
            log.debug("Program not found. Attempting to use cached version")

        #create c_src
        if recreate_c_src:
            with open(c_src_file, 'w') as c_file:
                c_file.write(program_text)
            log.info("file for generated C: %s", c_src_file)
            # syntax-highlight and print C program
            highlighted = highlight(program_text, 'c')
            log.info("generated C program: (((\n%s\n)))", highlighted)
            self.program_hash = program_hash


        #create ll_bc_file
        if recreate_so:
            # call clang to generate LLVM bitcode file
            log.debug('Regenerating so.')
            CC = ctree.CONFIG.get(self.config_target, 'CC')
            CFLAGS = ctree.CONFIG.get(self.config_target, 'CFLAGS')
            LDFLAGS = ctree.CONFIG.get(self.config_target, 'LDFLAGS')
            compile_cmd = "%s -shared %s -o %s %s %s" % (CC, CFLAGS, so_file,
                    c_src_file, LDFLAGS)
            log.info("compilation command: %s", compile_cmd)
            subprocess.check_call(compile_cmd, shell=True)
            # log.info("file for generated so: %s", so_file)

        #use cached version otherwise
        if not (so_file_exists or recreate_so):
            raise NotImplementedError('No Cached version found')


        # load llvm bitcode
        # import llvm.core
        # import llvmlite.binding as llvm

        # with open(ll_bc_file, 'rb') as bc:
        #     ll_module = llvm.module.parse_bitcode(bc.read())

        # syntax-highlight and print LLVM program
        #preserve_src_drhighlighted = highlight(str(ll_module), 'llvm')
        #log.debug("generated LLVM Program: (((\n%s\n)))", highlighted)

        return so_file


class MultiNode(CNode):
    """
        Some Python nodes need to be translated to a block of nodes but Visitors can't do that.
    """

    _fields = ['body']
    _requires_semicolon = lambda self: False

    def __init__(self, body = None):
        self.body = body or []
        CNode.__init__(self)


class Statement(CNode):
    """Section B.2.3 6.6."""
    pass


class Expression(CNode):
    """Cite me."""


class Return(Statement):
    """Section B.2.3 6.6.6 line 4."""
    _fields = ['value']

    def __init__(self, value=None):
        self.value = value
        super(Return, self).__init__()


class If(Statement):
    """Cite me."""
    _fields = ['cond', 'then', 'elze']

    def __init__(self, cond=None, then=None, elze=None):
        self.cond = cond
        self.then = then
        self.elze = elze
        super(If, self).__init__()


class While(Statement):
    """Cite me."""
    _fields = ['cond', 'body']
    _requires_semicolon = lambda self: False

    def __init__(self, cond=None, body=None):
        self.cond = cond
        self.body = body if body else []
        super(While, self).__init__()


class DoWhile(Statement):
    _fields = ['body', 'cond']

    def __init__(self, body=None, cond=None):
        self.body = body if body else []
        self.cond = cond
        super(DoWhile, self).__init__()


class For(Statement):
    _fields = ['init', 'test', 'incr', 'body']

    def __init__(self, init=None, test=None, incr=None, body=None):
        self.init = init
        self.test = test
        self.incr = incr
        self.body = body
        super(For, self).__init__()


# class Define(Statement):
# mbd: deprecated. see ctree.cpp.nodes.CppDefine


class FunctionCall(Expression):
    """Cite me."""
    _fields = ['func', 'args']

    def __init__(self, func=None, args=None):
        self.func = func
        self.args = args if args else []
        super(FunctionCall, self).__init__()


class Literal(Expression):
    """Cite me."""
    pass


class Constant(Literal):
    """Section B.1.4 6.1.3."""
    _fields = ['value']

    def __init__(self, value=None):
        self.value = value
        super(Constant, self).__init__()

    def get_type(self):
        return get_ctype(self.value)


class Block(Statement):
    """Cite me."""
    _fields = ['body']

    def __init__(self, body=None):
        self.body = body if body else []
        super(Block, self).__init__()

    def _requires_semicolon(self):
        return False


class String(Literal):
    """Cite me."""

    def __init__(self, *values):
        self.values = values
        super(String, self).__init__()


class SymbolRef(Literal):
    """Cite me."""
    _next_id = 0
    _fields = ['name','type']

    def __init__(self, name=None, sym_type=None, _global=False,
                 _local=False, _const=False):
        """
        Create a new symbol with the given name. If a declaration
        type is specified, the symbol is considered a declaration
        and unparsed with the type.
        """
        self.name = name
        self.type = sym_type
        self._global = _global
        self._local = _local
        self._const = _const
        super(SymbolRef, self).__init__()

    def set_global(self, value=True):
        self._global = value
        return self

    def set_local(self, value=True):
        self._local = value
        return self

    def set_const(self, value=True):
        self._const = value
        return self

    @classmethod
    def unique(cls, name="name", sym_type=None):
        """
        Factory for making unique symbols.
        """
        sym = SymbolRef("%s_%d" % (name, cls._next_id), sym_type)
        cls._next_id += 1
        return sym

    def copy(self, declare=False):
        if declare:
            return SymbolRef(self.name, self.type, self._global)
        else:
            return SymbolRef(self.name)

class FunctionDecl(Statement):
    """Cite me."""
    _fields = ['params', 'defn']

    def __init__(self, return_type=None, name=None, params=None, defn=None):
        self.return_type = return_type
        self.name = name
        self.params = params if params else []
        self.defn = defn if defn else []
        self.inline = False
        self.static = False
        self.kernel = False
        super(FunctionDecl, self).__init__()

    def get_type(self):
        type_sig = []

        # return type
        if self.return_type is None:
            type_sig.append(self.return_type)
        else:
            assert not isinstance(self.return_type, type), \
                "Expected a ctypes instance or None, got %s (%s)." % \
                    (self.return_type, type(self.return_type))
            type_sig.append( type(self.return_type) )

        # parameter types
        for param in self.params:
            assert not isinstance(param.type, type), \
                "Expected a ctypes instance or None, got %s (%s)." % \
                    (param.type, type(param.type))
            type_sig.append( type(param.type) )

        return CFUNCTYPE(*type_sig)

    def set_inline(self, value=True):
        self.inline = value
        return self

    def set_static(self, value=True):
        self.static = value
        return self

    def set_kernel(self, value=True):
        self.kernel = value
        return self


class UnaryOp(Expression):
    """Cite me."""
    _fields = ['arg']

    def __init__(self, op=None, arg=None):
        self.op = op
        self.arg = arg
        super(UnaryOp, self).__init__()


class BinaryOp(Expression):
    """Cite me."""
    _fields = ['left', 'op', 'right']

    def __init__(self, left=None, op=None, right=None):
        self.left = left
        self.op = op
        self.right = right
        super(BinaryOp, self).__init__()

    def get_type(self, env=None):
        # FIXME: integer promotions and stuff like that
        if hasattr(self.left, 'get_type'):
            left_type = self.left.get_type()
        elif isinstance(self.left, SymbolRef) and env is not None \
                and env._has_key(self.left.name):
            left_type = env._lookup(self.left.name)
        elif hasattr(self.left, 'type'):
            left_type = self.left.type
        else:
            left_type = None
        if hasattr(self.right, 'get_type'):
            right_type = self.right.get_type()
        elif isinstance(self.right, SymbolRef) and env is not None \
                and env._has_key(self.right.name):
            right_type = env._lookup(self.right.name)
        elif hasattr(self.right, 'type'):
            right_type = self.right.type
        else:
            right_type = None
        return get_common_ctype(filter(lambda x: x is not None, [right_type,
                                                                 left_type]))


class AugAssign(Expression):
    """Cite me."""
    _fields = ['target', 'value']

    def __init__(self, target=None, op=None, value=None):
        self.target = target
        self.op = op
        self.value = value
        super(AugAssign, self).__init__()


class TernaryOp(Expression):
    """Cite me."""
    _fields = ['cond', 'then', 'elze']

    def __init__(self, cond=None, then=None, elze=None):
        self.cond = cond
        self.then = then
        self.elze = elze
        super(TernaryOp, self).__init__()


class Cast(Expression):
    """doc"""
    _fields = ['value']

    def __init__(self, sym_type=None, value=None):
        self.type = sym_type
        self.value = value
        super(Cast, self).__init__()


class ArrayDef(Expression):
    """doc"""
    _fields = ['target', 'size', 'body']

    def __init__(self, target=None, size=None, body=None):
        self.target = target
        self.size = size
        self.body = body if body else []
        super(ArrayDef, self).__init__()

class Array(Expression):
    _fields = ['type', 'size', 'body']

    def __init__(self, type, size = None, body = None):
        self.body = body or []
        self.size = size or len(self.body)
        self.type = type
        super(Array, self).__init__()

    def get_type(self):
        return self.type

class Break(Statement):
    _requires_semicolon = lambda self : True

class Continue(Statement):
    _requires_semicolon = lambda self : True

class Pass(Statement):
    _requires_semicolon = lambda self: False


@singleton
class Op:
    class _Op(object):
        def __init__(self):
            self._force_parentheses = False

        def __str__(self):
            return self._c_str

    class PreInc(_Op):
        _c_str = "++"

    class PreDec(_Op):
        _c_str = "--"

    class PostInc(_Op):
        _c_str = "++"

    class PostDec(_Op):
        _c_str = "--"

    class Ref(_Op):
        _c_str = "&"

    class Deref(_Op):
        _c_str = "*"

    class SizeOf(_Op):
        _c_str = "sizeof"

    class Add(_Op):
        _c_str = "+"

    class AddUnary(_Op):
        _c_str = "+"

    class Sub(_Op):
        _c_str = "-"

    class SubUnary(_Op):
        _c_str = "-"

    class Mul(_Op):
        _c_str = "*"

    class Div(_Op):
        _c_str = "/"

    class Mod(_Op):
        _c_str = "%"

    class Gt(_Op):
        _c_str = ">"

    class Lt(_Op):
        _c_str = "<"

    class GtE(_Op):
        _c_str = ">="

    class LtE(_Op):
        _c_str = "<="

    class Eq(_Op):
        _c_str = "=="

    class NotEq(_Op):
        _c_str = "!="

    class BitAnd(_Op):
        _c_str = "&"

    class BitOr(_Op):
        _c_str = "|"

    class BitNot(_Op):
        _c_str = "~"

    class BitShL(_Op):
        _c_str = "<<"

    class BitShR(_Op):
        _c_str = ">>"

    class BitXor(_Op):
        _c_str = "^"

    class And(_Op):
        _c_str = "&&"

    class Or(_Op):
        _c_str = "||"

    class Not(_Op):
        _c_str = "!"

    class Comma(_Op):
        _c_str = ","

    class Dot(_Op):
        _c_str = "."

    class Arrow(_Op):
        _c_str = "->"

    class Assign(_Op):
        _c_str = "="

    class ArrayRef(_Op):
        _c_str = "[]"


# ---------------------------------------------------------------------------
# factory routines for building UnaryOps, BinaryOps, etc.

def PreInc(a):
    return UnaryOp(Op.PreInc(), a)


def PreDec(a):
    return UnaryOp(Op.PreDec(), a)


def PostInc(a):
    return UnaryOp(Op.PostInc(), a)


def PostDec(a):
    return UnaryOp(Op.PostDec(), a)


def BitNot(a):
    return UnaryOp(Op.BitNot(), a)


def Not(a):
    return UnaryOp(Op.Not(), a)


def Ref(a):
    return UnaryOp(Op.Ref(), a)


def Deref(a):
    return UnaryOp(Op.Deref(), a)


def SizeOf(a):
    return UnaryOp(Op.SizeOf(), a)


def Add(a, b=None):
    if b is not None:
        return BinaryOp(a, Op.Add(), b)
    else:
        return UnaryOp(Op.AddUnary(), a)


def Sub(a, b=None):
    if b is not None:
        return BinaryOp(a, Op.Sub(), b)
    else:
        return UnaryOp(Op.SubUnary(), a)


def Mul(a, b):
    return BinaryOp(a, Op.Mul(), b)


def Div(a, b):
    return BinaryOp(a, Op.Div(), b)


def Mod(a, b):
    return BinaryOp(a, Op.Mod(), b)


def Gt(a, b):
    return BinaryOp(a, Op.Gt(), b)


def Lt(a, b):
    return BinaryOp(a, Op.Lt(), b)


def GtE(a, b):
    return BinaryOp(a, Op.GtE(), b)


def LtE(a, b):
    return BinaryOp(a, Op.LtE(), b)


def Eq(a, b):
    return BinaryOp(a, Op.Eq(), b)


def NotEq(a, b):
    return BinaryOp(a, Op.NotEq(), b)


def BitAnd(a, b):
    return BinaryOp(a, Op.BitAnd(), b)


def BitOr(a, b):
    return BinaryOp(a, Op.BitOr(), b)


def BitShL(a, b):
    return BinaryOp(a, Op.BitShL(), b)


def BitShR(a, b):
    return BinaryOp(a, Op.BitShR(), b)


def BitXor(a, b):
    return BinaryOp(a, Op.BitXor(), b)


def And(a, b):
    return BinaryOp(a, Op.And(), b)


def Or(a, b):
    return BinaryOp(a, Op.Or(), b)


def Comma(a, b):
    return BinaryOp(a, Op.Comma(), b)


def Dot(a, b):
    return BinaryOp(a, Op.Dot(), b)


def Arrow(a, b):
    return BinaryOp(a, Op.Arrow(), b)


def Assign(a, b):
    return BinaryOp(a, Op.Assign(), b)


def ArrayRef(a, b):
    return BinaryOp(a, Op.ArrayRef(), b)


def AddAssign(a, b):
    return AugAssign(a, Op.Add(), b)


def SubAssign(a, b):
    return AugAssign(a, Op.Sub(), b)


def MulAssign(a, b):
    return AugAssign(a, Op.Mul(), b)


def DivAssign(a, b):
    return AugAssign(a, Op.Div(), b)


def ModAssign(a, b):
    return AugAssign(a, Op.Mod(), b)


def BitXorAssign(a, b):
    return AugAssign(a, Op.BitXor(), b)


def BitAndAssign(a, b):
    return AugAssign(a, Op.BitAnd(), b)


def BitOrAssign(a, b):
    return AugAssign(a, Op.BitOr(), b)


def BitShLAssign(a, b):
    return AugAssign(a, Op.BitShL(), b)


def BitShRAssign(a, b):
    return AugAssign(a, Op.BitShR(), b)



"""
Macros for working with SIMD vector types.
"""

from ctree.c.nodes import FunctionCall, SymbolRef

def _make_call(name, nArgs):
    def _the_call(*args):
        assert len(args) == nArgs, \
            "Macro expected %d args, got %d." % (nArgs, len(args))
        return FunctionCall(SymbolRef(name), args)
    return _the_call

mm256_storeu_pd = _make_call("_mm256_storeu_pd", 2)
mm256_loadu_pd  = _make_call("_mm256_loadu_pd", 1)
mm256_load_pd   = _make_call("_mm256_load_pd", 1)
mm256_set1_pd   = _make_call("_mm256_set1_pd", 1)
mm256_add_pd    = _make_call("_mm256_add_pd", 2)
mm256_mul_pd    = _make_call("_mm256_mul_pd", 2)



"""
Utilities for determining precedence in C, with the goal of minimizing the
number of parentheses in the generated code.
"""

from ctree.c.nodes import UnaryOp, BinaryOp, TernaryOp, Cast

# ---------------------------------------------------------------------------
# dictionary of node class -> precedence ids, where 1 denotes the highest-
# precedence operator and 18 is the lowest. This is how the table on
# wikipedia does it, but we flip it later so larger numbers mean higher
# precedence. For the origin of this table see
# http://en.wikipedia.org/wiki/Operators_in_C_and_C%2B%2B#Operator_precedence
_EXPR_TO_PRECEDENCE = {
    'PostInc': 2,
    'PostDec': 2,
    'FunctionCall': 2,
    'ArrayRef': 2,
    'Dot': 2,
    'Arrow': 2,

    'PreInc': 3,
    'PreDec': 3,
    #'Plus': 3,
    # TODO: work out naming schemes for add/sub unary
    'AddUnary': 3,
    #'Minus': 3,
    'SubUnary': 3,
    'Not': 3,
    'BitNot': 3,
    'Cast': 3,
    'Deref': 3,
    'Ref': 3,
    'SizeOf': 3,

    'Mul': 5,
    'Div': 5,
    'Mod': 5,

    'Add': 6,
    'Sub': 6,

    'BitShL': 7,
    'BitShR': 7,

    'Lt': 8,
    'LtE': 8,
    'Gt': 8,
    'GtE': 8,

    'Eq': 9,
    'NotEq': 9,

    'BitAnd': 10,

    'BitXor': 11,

    'BitOr': 12,

    'And': 13,

    'Or': 14,

    'TernaryOp': 15,

    'Assign': 16,
    'AddAssign': 16,
    'SubAssign': 16,
    'MulAssign': 16,
    'DivAssign': 16,
    'ModAssign': 16,
    'BitShLAssign': 16,
    'BitShRAssign': 16,
    'BitAndAssign': 16,
    'BitXorAssign': 16,
    # 'BitNotAssign': 16,

    'Comma': 18,
}


def get_precedence(node):
    try:
        if isinstance(node, (UnaryOp, BinaryOp)):
            op = type(node.op)
        elif isinstance(node, TernaryOp):
            op = TernaryOp
        elif isinstance(node, Cast):
            op = Cast
        else:
            raise Exception("Cannot determine operator for node %d.\n", node)

        pred = _EXPR_TO_PRECEDENCE[op.__name__]
        # flip precedence so higher numbers mean higher precedence
        return 20 - pred
    except KeyError:
        pass
    raise Exception("Unable to determine precedence for %s." % type(node).__name__)


_PRECEDENCE_ASSOCIATES_LTR = {
    2: True,
    3: False,
    5: True,
    6: True,
    7: True,
    8: True,
    9: True,
    10: True,
    11: True,
    12: True,
    13: True,
    14: True,
    15: False,
    16: False,
    18: True
}


def is_left_associative(node):
    try:
        pred = get_precedence(node)
        return _PRECEDENCE_ASSOCIATES_LTR[20 - pred]
    except (Exception, KeyError):
        pass
    raise Exception("Cannot determine if operator %s is left- or right-associative." % type(node).__name__)



import unittest
from ctypes import c_char_p

from ctree.nodes import *
from ctree.c.nodes import *

class TestPathRefs(unittest.TestCase):
    def test_self_ref(self):
        cfile = CFile("generated", [])
        tree = Project([cfile])
        stmt = Assign(SymbolRef("path", c_char_p()), \
                      cfile.get_generated_path_ref())
        cfile.body.append(stmt)

        proj = Project([cfile])
        llvm_module = proj.codegen() # triggers path resolution

        # self.assertIsNone( proj.find(GeneratedPathRef) )
        # self.assertIsNotNone( proj.find(String) )

    def test_other_ref(self):
        from ctree.ocl.nodes import OclFile

        cfile = CFile("generated", [])
        oclfile = OclFile("kernel", [])
        tree = Project([cfile])
        stmt = Assign(SymbolRef("path", c_char_p()), \
                      oclfile.get_generated_path_ref())
        cfile.body.append(stmt)

        proj = Project([cfile])
        llvm_module = proj.codegen() # triggers path resolution
        #
        # self.assertIsNone( proj.find(GeneratedPathRef) )
        # self.assertIsNotNone( proj.find(String) )



import types
import ctypes
import _ctypes
import sys

from ctree.types import (
    codegen_type,
    register_type_recognizers,
    register_type_codegenerators,
)

#Py2 and Py3 common types

register_type_recognizers(
    {
        int: ctypes.c_long,
        bool: ctypes.c_bool,
        float: ctypes.c_double,
        str: lambda t: ctypes.c_char(t.encode()) if len(t) == 1 else ctypes.c_char_p(t.encode()),
        type(None): lambda t: None
    }
)

register_type_codegenerators({
    ctypes.c_int: lambda t: "int",
    ctypes.c_long: lambda t: "long",
    ctypes.c_float: lambda t: "float",
    ctypes.c_double: lambda t: "double",
    ctypes.c_char: lambda t: "char",
    ctypes.c_char_p: lambda t: "char*",
    ctypes.c_void_p: lambda t: "void*",
    ctypes.c_bool: lambda t: "bool",
    ctypes.c_ulong: lambda t: "size_t",
    type(None): lambda n: "void",

    _ctypes.Array: lambda ct: "%s*" % codegen_type(ct._type_()),
    _ctypes._Pointer: lambda ct: "%s*" % codegen_type(ct._type_()),

})

#register version specific nodes

if sys.version_info >= (3, 0):
    pass

else:

    register_type_recognizers({
        long: ctypes.c_long
    })



"""
Code generator for C constructs.
"""

from ctree.codegen import CodeGenVisitor
from ctree.c.nodes import Op
from ctree.types import codegen_type
from ctree.precedence import UnaryOp, BinaryOp, TernaryOp, Cast
from ctree.precedence import get_precedence, is_left_associative

from ctree.nodes import CommonCodeGen

class CCodeGen(CommonCodeGen):
    """
    Manages generation of C code.
    """

    def _requires_parentheses(self, parent, node):
        """
        Returns node as a string, optionally with parentheses around it if
        needed to enforce precendence rules.
        """
        if isinstance(node, (UnaryOp, BinaryOp, TernaryOp)) and\
                isinstance(parent, (UnaryOp, BinaryOp, TernaryOp, Cast)):
            prec = get_precedence(node)
            parent_prec = get_precedence(parent)
            is_not_last_child = isinstance(parent, UnaryOp) or\
                                isinstance(parent, Cast) or\
                                (isinstance(parent, BinaryOp) and node is parent.left) or\
                                (isinstance(parent, TernaryOp) and node is not parent.elze)
            assoc_left = is_left_associative(parent)
            if (prec < parent_prec) or \
                    (prec == parent_prec and (assoc_left is not is_not_last_child)):
                return True
        return False

    # -------------------------------------------------------------------------
    # visitor methods

    def visit_MultiNode(self, node):
        return self._genblock(node.body, insert_curly_brackets=False, increase_indent=False)

    def visit_FunctionDecl(self, node):
        params = ", ".join(map(str, node.params))
        s = ""
        if node.kernel:
            s += "__kernel "
        if node.static:
            s += "static "
        if node.inline:
            s += "inline "
        s += "%s %s(%s)" % (codegen_type(node.return_type), node.name, params)
        if node.defn:
            s += " %s" % self._genblock(node.defn)
        return s

    def visit_UnaryOp(self, node):
        op  = self._parenthesize(node, node.op)
        arg = self._parenthesize(node, node.arg)
        if isinstance(node.op, (Op.PostInc, Op.PostDec)):
            return "%s %s" % (arg, op)
        else:
            return "%s %s" % (op, arg)

    def visit_BinaryOp(self, node):
        left  = self._parenthesize(node, node.left)
        right = self._parenthesize(node, node.right)
        if isinstance(node.op, Op.ArrayRef):
            return "%s[%s]" % (left, right)
        else:
            return "%s %s %s" % (left, node.op, right)

    def visit_AugAssign(self, node):
        return "%s %s= %s" % (node.target, node.op, node.value)

    def visit_TernaryOp(self, node):
        cond = self._parenthesize(node, node.cond)
        then = self._parenthesize(node, node.then)
        elze = self._parenthesize(node, node.elze)
        return "%s ? %s : %s" % (cond, then, elze)

    def visit_Cast(self, node):
        value = self._parenthesize(node, node.value)
        return "(%s) %s" % (codegen_type(node.type), value)

    def visit_Constant(self, node):
        if isinstance(node.value, str):
            return "'%s'" % node.value[0]
        else:
            return str(node.value)

    def visit_SymbolRef(self, node):
        s = ""
        if node._global:
            s += "__global "
        if node._local:
            s += "__local "
        if node._const:
            s += "const "
        if node.type is not None:
            s += "%s " % codegen_type(node.type)
        return "%s%s" % (s, node.name)

    def visit_Block(self, node):
        return self._genblock(node.body)

    def visit_Return(self, node):
        if node.value:
            return "return %s" % node.value
        else:
            return "return"

    def visit_If(self, node):
        then = self._genblock(node.then)
        if node.elze:
            elze = self._genblock(node.elze)
            return "if (%s) %s else %s" % (node.cond, then, elze)
        else:
            return "if (%s) %s" % (node.cond, then)

    def visit_While(self, node):
        body = self._genblock(node.body)
        return "while (%s) %s" % (node.cond, body)

    def visit_DoWhile(self, node):
        body = self._genblock(node.body)
        return "do %s while (%s)" % (body, node.cond)

    def visit_For(self, node):
        body = self._genblock(node.body)
        return "for (%s; %s; %s) %s" % (node.init, node.test, node.incr, body)

    def visit_FunctionCall(self, node):
        args = ", ".join(map(str, node.args))
        return "%s(%s)" % (node.func, args)

    def visit_String(self, node):
        return '"%s"' % '" "'.join(node.values)

    def visit_CFile(self, node):
        stmts = self._genblock(node.body, insert_curly_brackets=False, increase_indent=False)
        return '// <file: %s>%s' % (node.get_filename(), stmts)

    def visit_ArrayDef(self, node):
        body = ", ".join(map(str, node.body))
        return "%s[%s] = { %s }" % (node.target, node.size, body)

    def visit_Break(self, node):
        return 'break'

    def visit_Continue(self, node):
        return 'continue'

    def visit_Array(self, node):
        return "{%s}" % ', '.join([i.codegen() for i in node.body])




#from distutils.core import setup
from setuptools import setup


def make_data_file_list(target, source):
    """
    creates a list of tuples (target_directory, file_list)
    that setup should copy into the egg in
    the install process.  Gets around no way to specify a recursive
    directory tree copy, directories must be created by
    name on left side of tuple against an empty file list on the right
    """
    import os

    def visit(destination_directory, source_directory):
        # print destination_directory + " " + source_directory
        files = os.listdir(source_directory)
        transfer_list = []
        todo = []
        for file_name in files:
            # print type(file_name)
            source_file = os.path.join(source_directory, file_name)
            # print "source_file " + source_file
            next_destination_directory = os.path.join(destination_directory, file_name)

            if file_name.startswith("__") and file_name != '__init__.py':
                pass
            elif os.path.isdir(source_file):
                transfer_list.append((next_destination_directory, []))
                todo.append((next_destination_directory, source_file))
            elif os.path.isfile(source_file):
                transfer_list.append((destination_directory, [source_file]))

        for next_target, next_source in todo:
            transfer_list += visit(next_target, next_source)

        return transfer_list

    return visit(target, source)

data_file_list = make_data_file_list("ctree/tools/generators/templates", "ctree/tools/generators/templates")
# import pprint
# pprint.pprint(data_file_list)


setup(
    name='ctree',
    version='0.1.7',
    description='A C-family AST implementation designed to be an IR for DSL compilers.',

    packages=[
        'ctree',
        'ctree.c',
        'ctree.cilk',
        'ctree.cpp',
        'ctree.ocl',
        'ctree.omp',
        'ctree.py',
        'ctree.np',
        'ctree.simd',
        'ctree.templates',
        'ctree.opentuner',
        'ctree.metrics',
        'ctree.tools',
        'ctree.tools.generators',
        'ctree.tools.generators.templates',
        'ctree.transforms',
        'ctree.visual',
    ],

    package_data={
        'ctree': ['defaults.cfg'],
    },

    install_requires=[
        'numpy',
        'pyserial',
        'pycl'
    ],

    data_files=data_file_list,

    entry_points={
        'console_scripts': ['ctree = ctree.tools.runner:main'],
    }
)



__author__ = 'nzhang-dev'

import ast

from ctree.transformations import DeclarationFiller, PyBasicConversions
from ctree.frontend import *
from ctree.c.nodes import MultiNode


code = [
    "a = 1",
    "a,b = 1,1",
    "a = b = 1",
    """a,b = 1,1 \na,b = b,a"""
]

def fib(n):
    a,b = 0, 1
    while n > 0:
        n -= 1
        a, b = b, a+b
    return a

asts = []
for c in code:
    parsed = ast.parse(c)
    asts.append(MultiNode(body = parsed.body))

asts.append(get_ast(fib).body[0])

processed = [
    DeclarationFiller().visit(PyBasicConversions().visit(a)) for a in asts
]


"""
Reader that can sample data from a WattsUpMeter: https://www.wattsupmeters.com
API documents are here: https://www.wattsupmeters.com/secure/downloads/CommunicationsProtocol090824.pdf

basic usage is as follows

    reader = WattsUpReader("/dev/tty.usbserial-A600KI7M")
    reader.drain()

    reader.start_recording()

    #
    # do a bunch of stuff that uses power
    #

    summary, detailed_results = reader.get_recording()
    #
    # summary is a Summary named-tuple
    # detailed_results an array of Result named-tuples defined below
    #

This also is runnable from the command line

python watts_up_reader.py "/dev/tty.usbserial-A600KI7M"

which will put you in an interactive session, that allows testing
of recording, a few other things and direct communication with the
device

the meter is finicky and may not be working if connection is made
quickly after

"""
from __future__ import print_function

__author__ = 'Chick Markley'


import serial
import readline
import time
import select
import collections
import argparse


class WattsUpReader(object):
    EXTERNAL_MODE = "E"
    INTERNAL_MODE = "I"
    FULL_HANDLING = 2

    Result = collections.namedtuple("Result", ['time', 'watts', 'volts', 'milliamps'])
    Response = collections.namedtuple("Response", ['time', 'message'])
    Summary = collections.namedtuple(
        "Summary", ["joules", "millicoulomb", "samples", "sampling_interval", "start_time"]
    )

    def __init__(self, port_name=None, verbose=False):
        if port_name is None:
            from ctree import CONFIG
            port_name = CONFIG.get('wattsup', 'port')
        self.port_name = port_name
        self.serial_port = serial.Serial(self.port_name, 115200)

        self.last_time = time.time()
        self.verbose = verbose
        self.t = []
        self.power = []
        self.returned_lines = 0
        # I don't think device supports anything smaller
        self.record_interval = 1

        self.start_recording()

    def reset(self):
        if self.serial_port:
            self.serial_port.close()
        time.sleep(1)

        self.serial_port = serial.Serial(self.port_name, 115200)
        if self.verbose:
            print("serial port:")
            print(self.serial_port)
        self.serial_port.sendBreak()
        self.serial_port.flushInput()
        self.serial_port.flushOutput()
        self.serial_port.setDTR()
        self.serial_port.write(chr(0x18))
        time.sleep(1)
        self.send_command("#V,W,0;", timeout=10, tries=1)

    def clear(self):
        self.serial_port.write("#R,W,0;")
        self.drain()

    def set_verbose(self, new_value=None):
        if new_value is None:
            self.verbose = not self.verbose
        else:
            self.verbose = new_value

    def set_mode(self, runmode):
        """
        TODO: Fix this, ported from https://github.com/kjordahl/Watts-Up--logger
        See API url above

        """
        self.serial_port.write('#L,W,3,%s,,%d;' % (runmode, self.record_interval))
        if runmode == WattsUpReader.INTERNAL_MODE:
            self.serial_port.write('#O,W,1,%d' % WattsUpReader.FULL_HANDLING)

    def fetch(self, base_time=None, time_out=0.3, raw=False):
        """read one data point from meter"""

        rfds, wfds, efds = select.select([self.serial_port], [], [], time_out)
        if rfds:
            # device_output = self.serial_port.readline()..decode("utf-8")  #python3
            device_output = self.serial_port.readline()
            self.returned_lines += 1
            if self.verbose:
                print(device_output)
            if device_output.startswith("#d"):
                if raw:
                    return "%s,%s" % (device_output.strip(), base_time)

                fields = device_output.split(',')
                if self.verbose:
                    for index, field in enumerate(fields):
                        print("%02d %s" % (index, field))
                watts = float(fields[3]) / 10
                volts = float(fields[4]) / 10
                milliamps = float(fields[5]) / 1000
                if not base_time:
                    base_time = time.time()
                return WattsUpReader.Result(base_time, watts, volts, milliamps)
            elif len(device_output) > 0:
                return WattsUpReader.Response(time.time(), device_output)
        return None

    def drain(self):
        while True:
            result = self.fetch()
            if type(result) is WattsUpReader.Response:
                print(result)
            elif self.verbose:
                print(result)
            if not result:
                return

    def dump(self):
        while True:
            result = self.fetch(base_time=time.time(), time_out=1000)
            print(result)

    def raw_dump(self):
        while True:
            result = self.fetch(base_time=time.time(), time_out=1000, raw=True)
            print(result)

    def send_command(self, command, timeout=3, tries=3):
        if not command.startswith("#"):
            print( "Error: no initial # for command %s", command)
            return
        if not command.strip().endswith(";"):
            print( "Error: no trailing ; for command %s", command)
            return

        for tries in range(tries):
            if self.verbose:
                print("sending command %s" % command)
            self.serial_port.write(command)

            answer = self.fetch(time_out=timeout)
            if answer:
                if self.verbose:
                    print("answer %s",end="")
                    print(answer)
                self.last_time = time.time()
                return
            else:
                if self.verbose:
                    print("timed out sending command %s" % command)

    def start_recording(self):
        self.drain()

        command = "#L,W,3,E,,%d;" % self.record_interval
        self.send_command(command)

    def get_recording(self):
        def pull():
            results = []
            estimated_record_time = self.last_time
            watt_seconds = 0.0
            samples = 0
            millicoulombs = 0.0

            while True:
                result = self.fetch(estimated_record_time)
                if not result:
                    summary = WattsUpReader.Summary(
                        watt_seconds, millicoulombs, samples, self.record_interval, self.last_time
                    )
                    return summary, results
                if type(result) is WattsUpReader.Result:
                    results.append(result)
                    watt_seconds += result.watts
                    millicoulombs += result.milliamps
                    samples += 1
                    estimated_record_time += self.record_interval
        all_results = pull()
        self.last_time = time.time()
        return all_results

    def command(self, string):
        self.serial_port.write(string)
        save_verbose, self.verbose = self.verbose, True
        time.sleep(1)
        self.drain()
        self.verbose = save_verbose

    @staticmethod
    def usage():
        print("command must be one of (quit, reset, record, get_record, verbose, drain) or ")
        print("native device command string beginning with # ")
        print("empty command will repeat previous command")
        print("commands can be abbreviated to first three letters")
        print("\n")

    def interactive_mode(self):
        readline.parse_and_bind('tab: complete')
        readline.parse_and_bind('set editing-mode vi')

        last_input = None
        WattsUpReader.usage()
        while True:
            # print("Command: ", end="")
            # user_input = sys.stdin.readline()

            user_input = raw_input("Command (quit to exit): ")
            # print("got input %s" % user_input)
            # if user_input.strip() == '':
            #     user_input = last_input

            if user_input.startswith('q') or user_input.startswith('Q'):
                return
            elif user_input.lower().startswith('res'):
                self.reset()
            elif user_input.lower().startswith('dra'):
                self.drain()
            elif user_input.lower().startswith('rec'):
                self.start_recording()
            elif user_input.lower().startswith('get'):
                summary, detailed_results = self.get_recording()
                for index, result in enumerate(detailed_results):
                    print("%d ----" % index, end='')
                    print(result)
                print(summary)
            elif user_input.lower().startswith('ver'):
                self.verbose = not self.verbose
            elif user_input.startswith("#"):
                print("sending...")
                self.serial_port.write(user_input)
                time.sleep(1)
                self.drain()
            else:
                print("unknown command: %s" % user_input)
                WattsUpReader.usage()
            last_input = user_input

    def stop(self):
        self.serial_port.close()

    @staticmethod
    def guess_port():
        import subprocess

        for tty_search_command in ["ls /dev/tty*usb*","ls /dev/tty*USB*"]:
            try:
                possible_devices = subprocess.check_output(tty_search_command, shell=True).strip().split('\n')
            except:
                possible_devices = []

            if len(possible_devices) == 1:
                return possible_devices[0]
            else:
                for device in possible_devices:
                    print("Possible device %s" % device)
                print("Multiple possible devices found, you must specify explicitly")
                exit(1)

        print("No potential usb based readers found, is it plugged in?")
        exit(1)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="interface to WattsUpPro usb power meter")
    parser.add_argument(
        '-i', '--interactive', help='interactive mode, allows direct communcation with device', action="store_true"
    )
    parser.add_argument('-p', '--port', help='full /dev/ path to the usb device, if missing a good quess will be made')
    parser.add_argument('-d', '--dump', help='meter data to stdout, default is prettified', action="store_true")
    parser.add_argument('-r', '--raw', help='modify dump to return raw device output', action="store_true")
    parser.add_argument('-c', '--clear', help='clear any saved lines in device', action="store_true")
    parser.add_argument('-v', '--verbose', help='show more debug than you like', action="store_true")

    args = parser.parse_args()

    if not args.port:
        usb_port_name = WattsUpReader.guess_port()
        print("Using port %s" % usb_port_name)
    else:
        usb_port_name = args.port

    watt_reader = WattsUpReader(usb_port_name, verbose=False)

    if args.verbose:
        watt_reader.set_verbose(True)
    if args.clear:
        watt_reader.drain()

    if args.interactive:
        watt_reader.interactive_mode()
    if args.raw:
        watt_reader.raw_dump()
    elif args.dump:
        watt_reader.dump()
    else:
        parser.print_usage()

    watt_reader.stop()



import unittest
from inspect import getsource

from ctree.frontend import *
from fixtures.sample_asts import *


class TestFrontend(unittest.TestCase):
    def test_identity(self):
        self.assertIsInstance(get_ast(identity), ast.AST)

    def test_gcd(self):
        self.assertIsInstance(get_ast(gcd), ast.AST)

    def test_fib(self):
        self.assertIsInstance(get_ast(fib), ast.AST)

    def test_parse_print(self):
        parseprint(getsource(fib))

    def test_dump(self):
        dump(fib_ast)
        #self.assertEqual(dump(fib_ast), 'FunctionDecl(params=[\n    SymbolRef(),\n  ], defn=[\n    If(cond=BinaryOp(left=SymbolRef(), right=Constant()), then=[\n        Return(value=SymbolRef()),\n      ], elze=[\n        Return(value=BinaryOp(left=FunctionCall(func=SymbolRef(), args=[\n            BinaryOp(left=SymbolRef(), right=Constant()),\n          ]), right=FunctionCall(func=SymbolRef(), args=[\n            BinaryOp(left=SymbolRef(), right=Constant()),\n          ]))),\n      ]),\n  ])')


"""
OpenMP nodes supported by ctree.
"""

import logging

log = logging.getLogger(__name__)

from ctree.nodes import CtreeNode

# ---------------------------------------------------------------------------
# openmp nodes


class OmpNode(CtreeNode):
    """Base class for all OpenMP nodes supported by ctree."""

    def codegen(self, indent=0):
        from ctree.omp.codegen import OmpCodeGen

        return OmpCodeGen(indent).visit(self)

    def label(self):
        from ctree.omp.dotgen import OmpDotLabeller

        return OmpDotLabeller().visit(self)

    def _requires_semicolon(self):
        return False


class OmpParallel(OmpNode):
    """
    Represents '#pragma omp parallel' annotations.
    """
    _fields = ['clauses']

    def __init__(self, clauses=None):
        self.clauses = clauses if clauses else []


class OmpParallelFor(OmpNode):
    """ #pragma omp parallel for ... """
    _fields = ['clauses']

    def __init__(self, clauses=None):
        self.clauses = clauses if clauses else []


class OmpParallelSections(OmpNode):
    """ #pragma omp parallel sections... """
    _fields = ['clauses', 'sections']

    def __init__(self, clauses=None, sections=None):
        self.clauses = clauses or []
        self.sections = sections or []


class OmpSection(OmpNode):
    """ #pragma omp section ... """
    _fields = ['clauses', 'body']

    def __init__(self, clauses=None, body=None):
        self.clauses = clauses or []
        self.body = body or []


class OmpIvDep(OmpNode):
    _field = ['clauses']

    def __init__(self, clauses=[]):
        self.clauses = clauses


class OmpClause(OmpNode):
    """Base class for OpenMP clauses."""
    pass


class OmpIfClause(OmpClause):
    _fields = ["exp"]

    def __init__(self, exp=None):
        self.exp = exp


class OmpNumThreadsClause(OmpClause):
    _fields = ["val"]

    def __init__(self, val=None):
        self.val = val


class OmpNoWaitClause(OmpClause):
    pass



import logging
log = logging.getLogger(__name__)

import abc
import itertools

class TuningDriver(object):
    """
    Object that interacts with backend tuners. Provides
    an infinite stream of configurations, as well as an
    interface to report on the performance of each.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        """
        Creates a tuner to search in the given space.
        """
        self.configs = self._get_configs()

    @abc.abstractmethod
    def _get_configs(self):
        """A generator that yields a stream of configurations."""
        pass

    @abc.abstractmethod
    def report(self, **kwargs):
        """Reports performance of most recent configuration."""
        pass


class ConstantTuningDriver(TuningDriver):
    """
    Provides a stream of the same config, and ignores reports()s.
    """
    def __init__(self, config=None):
        """Do nothing."""
        super(ConstantTuningDriver, self).__init__()
        self._config = config

    def _get_configs(self):
        """Yield the empty configuration."""
        while True:
            yield self._config

    def report(self, *args, **kwargs):
        """Ignore reports."""
        pass



class Parameter(object):
    """A dimension of the search space."""
    def __init__(self, name):
        """Create a parameter with the given name."""
        self.name = name
        self._values = []

    def values(self):
        return self._values


class IntegerParameter(Parameter):
    """An integer parameter."""
    def __init__(self, name, lower_bound, upper_bound):
        """Create an int parameter with value in range(lb, ub)."""
        super(IntegerParameter, self).__init__(name)
        self._values = range(lower_bound, upper_bound)


class BooleanParameter(Parameter):
    """A boolean parameter."""
    def __init__(self, name):
        """Create a bool parameter."""
        super(BooleanParameter, self).__init__(name)
        self._values = [True, False]


class EnumParameter(Parameter):
    """A enum parameter."""
    def __init__(self, name, values):
        """Create an enum parameter."""
        super(EnumParameter, self).__init__(name)
        self._values = values


class IntegerArrayParameter(Parameter):
    """An array of integers."""
    def __init__(self, name, count=1, lower_bound=0, upper_bound=1):
        """Create an IntArray parameter."""
        super(IntegerArrayParameter, self).__init__(name)
        self._values = itertools.product(range(lower_bound,upper_bound), repeat=count)


class EnumArrayParameter(Parameter):
    """An array of enums."""
    def __init__(self, name, count=1, values=None):
        """Create an EnumArray parameter."""
        super(EnumArrayParameter, self).__init__(name)
        values = values if values else []
        self._values = itertools.product(values, repeat=count)


class BooleanArrayParameter(Parameter):
    """An array of booleans."""
    def __init__(self, name, count=1):
        """Create an BooleanArray parameter."""
        super(BooleanArrayParameter, self).__init__(name)
        self._values = itertools.product([True,False], repeat=count)


class Result(object):
    """
    Captures the performance of a tuning run.
    """
    def __init__(self, time       = float('inf'),
                       accuracy   = float('-inf'),
                       energy     = float('inf'),
                       size       = float('inf'),
                       confidence = float('-inf')):
        self.time = time
        self.accuracy = accuracy
        self.energy = energy
        self.size = size
        self.confidence = confidence


class Objective(object):
    def compare(self, result0, result1):
        raise NotImplementedException()


class MinimizeTime(Objective):
    def compare(self, result0, result1):
        return result0.time < result1.time


class BruteForceTuningDriver(TuningDriver):
    """
    Yields all points in the hyperrectangular space defined by
    the cartesian product of all input parameters. Then, yields
    the best point for the rest of time.
    """
    def __init__(self, params, objective):
        """Initialize with the given objective."""
        super(BruteForceTuningDriver, self).__init__()
        self._objective = objective
        self._best_result = Result()
        self._best_cfg = None
        self._last_cfg = None
        self._params = params

    def _get_configs(self):
        """A generator of all configurations."""
        # compute the cartesian product of the dimensions
        names, values = zip(*[(p.name, p.values()) for p in self._params])
        for cfg_values in itertools.product(*values):
            cfg = {k: v for k,v in zip(names, cfg_values)}
            self._last_cfg = cfg
            yield cfg

        assert self._best_cfg != None, \
            "No best configuration. Did you call report()?"

        # return the best for the rest of time
        while True:
            yield self._best_cfg

    def report(self, **kwargs):
        """
        Record the new result if it is better than the current best.
        """
        new_result = Result(**kwargs)
        if self._objective.compare(new_result, self._best_result):
            self._best_result = new_result
            self._best_cfg = self._last_cfg



import unittest

from ctree.c.nodes import *


class TestTernaryOps(unittest.TestCase):
    def setUp(self):
        self.foo = SymbolRef("foo")
        self.bar = SymbolRef("bar")
        self.baz = SymbolRef("baz")

    def test_ternary_op(self):
        node = TernaryOp(self.foo, self.bar, self.baz)
        self.assertEqual(str(node), "foo ? bar : baz")



"""
SIMD vector nodes supported by ctree.
"""

from ctree.nodes import CtreeNode


class SimdNode(CtreeNode):
    """Base class for all SIMD nodes supported by ctree."""

    def codegen(self, indent=0):
        from ctree.sse.codegen import SimdCodeGen

        return SimdCodeGen(indent).visit(self)

    def label(self):
        from ctree.sse.dotgen import SimdDotLabeller

        return SimdDotLabeller().visit(self)



import ctypes as ct

from util import CtreeTest
from ctree.c.nodes import SymbolRef, Constant, Add, Mul, ArrayDef, Sub


class TestArrayDefs(CtreeTest):

    def test_simple_array_def(self):
        self._check_code(ArrayDef(
            SymbolRef('hi', ct.c_int()), Constant(2),
            [Constant(0), Constant(1)]
        ), "int hi[2] = { 0, 1 }")

    def test_complex(self):
        node = ArrayDef(
            SymbolRef('myArray', ct.c_int()),
            Constant(2),
            [
                Add(SymbolRef('b'), SymbolRef('c')),
                Mul(Sub(Constant(99), SymbolRef('d')), Constant(200))
            ]
        )
        self._check_code(node, "int myArray[2] = { b + c, (99 - d) * 200 }")



"""
Code generation for OpenMP.
"""

from ctree.codegen import CodeGenVisitor


class OmpCodeGen(CodeGenVisitor):
    """
    Visitor to generate omp code.
    """

    def visit_OmpParallel(self, node):
        s = "#pragma omp parallel"
        if node.clauses:
            s += " " + ", ".join(map(str, node.clauses))
        return s

    def visit_OmpParallelFor(self, node):
        s = "#pragma omp parallel for"
        if node.clauses:
            s += " " + ", ".join(map(str, node.clauses))
        return s

    def visit_OmpParallelSections(self, node):
        s = "#pragma omp parallel sections"
        if node.clauses:
            s += " " + ", ".join(map(str, node.clauses))
        s += "\n%s%s" % (self._tab(), self._genblock(node.sections))
        return s

    def visit_OmpSection(self, node):
        s = "#pragma omp section"
        if node.clauses:
            s += " " + ", ".join(map(str, node.clauses))
        s += "\n%s%s" % (self._tab(), self._genblock(node.body))
        return s

    def visit_OmpIfClause(self, node):
        return "if(%s)" % node.exp

    def visit_OmpNumThreadsClause(self, node):
        return "num_threads(%s)" % node.val

    def visit_OmpNoWaitClause(self, node):
        return "nowait"

    def visit_OmpIvDep(self, node):
        s = "#pragma IVDEP"
        if node.clauses:
          s += " " + ", ".join(map(str, node.clauses))
        return s     



import unittest

from ctree.cpp.nodes import *
from ctree.c.nodes import Add, SymbolRef, Constant


class TestCppIncludes(unittest.TestCase):
    def test_include_angled(self):
        tree = CppInclude("foo.h")
        self.assertEqual(str(tree), "#include <foo.h>")

    def test_include_quotes(self):
        tree = CppInclude("foo.h", angled_brackets=False)
        self.assertEqual(str(tree), '#include "foo.h"')

class TestDefines(unittest.TestCase):
    def _check(self, node, expected_string):
        self.assertEqual(str(node), expected_string)

    def test_simple_macro(self):
        d1, d2 = SymbolRef("d1"), SymbolRef("d2")
        node = CppDefine("test_macro", [d1, d2], Add(d1, d2))
        self._check(node, "#define test_macro(d1, d2) (d1 + d2)")

    def test_no_args(self):
        node = CppDefine("test_macro", [], Constant(39))
        self._check(node, "#define test_macro() (39)")



"""
Code generation for template nodes.
"""

from ctree.codegen import CodeGenVisitor
from ctree.util import flatten


class TemplateCodeGen(CodeGenVisitor):
    """
    Visitor to generate templated code.
    """
    def _resolve(self, value):
        """
        Generated the string for the given child, which might be a
        list of lists (...of lists...) of nodes. For a single node,
        don't append a semicolon, otherwise append semicolons and
        newlines to each statement if they require it.
        """
        children = list(flatten(value))
        if len(children) == 1:
            return children[0].codegen()
        else:
            body = ""
            for child in children:
                semicolon_opt = ";" if child._requires_semicolon() else ""
                body += self._tab() + child.codegen() + semicolon_opt + "\n"
            return body

    def visit_TemplateNode(self, node):
        """Convert this template to a string."""
        resolved_children = {key: self._resolve(value) for key, value in node._children.items()}
        resolved_txt = node._template.safe_substitute(resolved_children)
        indented_text = ("\n" + self._tab()).join(resolved_txt.splitlines())
        return indented_text

    def visit_StringTemplate(self, node):
        """Convert this StringTemplate to a string."""
        return self.visit_TemplateNode(node)

    def visit_FileTemplate(self, node):
        """Convert this FileTemplate to a string."""
        return self.visit_TemplateNode(node)



import _ctypes

import numpy as np

from ctree.types import (
    get_ctype,
)
from util import CtreeTest
from ctree.c.nodes import SymbolRef

class TestTypeRecognizer(CtreeTest):
    def test_int_array(self):
        ty = get_ctype(np.arange(10, dtype=np.int32))
        self.assertIsInstance(ty, _ctypes.Array)

class TestTypeCodeGen(CtreeTest):
    def test_int_array_1d(self):
        ty = get_ctype(np.arange(10, dtype=np.int32))
        tree = SymbolRef("i", ty)
        self._check_code(tree, "int* i")

    def test_int_array_2d(self):
        ty = get_ctype(np.arange(10, dtype=np.int32).reshape(2,5))
        tree = SymbolRef("i", ty)
        self._check_code(tree, "int** i")



import unittest
import ctypes as ct

from ctree.c.nodes import *


class TestCtreeNode(unittest.TestCase):

    def test_bad_override_codegen(self):
        class BadNode(CtreeNode):
            pass
        with self.assertRaises(Exception):
            BadNode().codegen()

    def test_bad_override_to_dot(self):
        class BadNode(CtreeNode):
            pass
        with self.assertRaises(Exception):
            BadNode()._to_dot()

    def test_find_all_attr_error(self):
        tree = Add(SymbolRef('a'), Constant(10))
        try:
            tree.find_all(SymbolRef, type=ct.c_int())
        except AttributeError:
            self.fail("find_all should not raise AttributeError")



import unittest

from ctree.jit import *
from ctree import CONFIG
from fixtures.sample_asts import *
import ctypes
from ctree.transformations import PyBasicConversions

class TestTranslator(LazySpecializedFunction):
    def args_to_subconfig(self, args):
        return {'arg_typesig': tuple(type(get_ctype(a)) for a in args)}

    def transform(self, tree, program_config):
        arg_types = program_config[0]['arg_typesig']
        tree = PyBasicConversions().visit(tree.body[0])
        tree.return_type = arg_types[0]()
        for param, ty in zip(tree.params, arg_types):
            param.type = ty()
        return [CFile(tree.name, [tree])]

    def finalize(self, transform_result, program_config):
        proj = Project(transform_result)
        cfile = transform_result[0]
        arg_types = program_config[0]['arg_typesig']

        func_type = ctypes.CFUNCTYPE(arg_types[0], *arg_types)
        return BasicFunction(cfile.name, proj, func_type)

class BasicFunction(ConcreteSpecializedFunction):
    def __init__(self, entry, tree, typesig):
        self._c_function = self._compile(entry, tree, typesig)

    def __call__(self, *args, **kwargs):
        return self._c_function(*args, **kwargs)


class TestJit(unittest.TestCase):
    def test_identity(self):
        mod = JitModule()
        submod = CFile("test_identity", [identity_ast], path=CONFIG.get('jit','COMPILE_PATH')). \
            _compile(identity_ast.codegen())
        mod._link_in(submod)
        c_identity_fn = mod.get_callable(identity_ast.name,
                                         identity_ast.get_type())
        self.assertEqual(identity(1), c_identity_fn(1))
        self.assertEqual(identity(12), c_identity_fn(12))
        self.assertEqual(identity(123), c_identity_fn(123))

    def test_fib(self):
        mod = JitModule()
        submod = CFile("test_fib", [fib_ast], path=CONFIG.get('jit','COMPILE_PATH'))._compile(fib_ast.codegen())
        mod._link_in(submod)
        c_fib_fn = mod.get_callable(fib_ast.name,
                                    fib_ast.get_type())
        self.assertEqual(fib(1), c_fib_fn(1))
        self.assertEqual(fib(6), c_fib_fn(6))

    def test_gcd(self):
        mod = JitModule()
        submod = CFile("test_gcd", [gcd_ast], path=CONFIG.get('jit','COMPILE_PATH'))._compile(gcd_ast.codegen())
        mod._link_in(submod)
        c_gcd_fn = mod.get_callable(gcd_ast.name,
                                    gcd_ast.get_type())
        self.assertEqual(gcd(44, 122), c_gcd_fn(44, 122))
        self.assertEqual(gcd(27, 39), c_gcd_fn(27, 39))

    def test_choose(self):
        mod = JitModule()
        submod = CFile("test_choose", [choose_ast], path=CONFIG.get('jit','COMPILE_PATH')). \
            _compile(choose_ast.codegen())
        mod._link_in(submod)
        c_choose_fn = mod.get_callable(choose_ast.name,
                                       choose_ast.get_type())
        self.assertEqual(choose(0.2, 44, 122), c_choose_fn(0.2, 44, 122))
        self.assertEqual(choose(0.8, 44, 122), c_choose_fn(0.8, 44, 122))
        self.assertEqual(choose(0.3, 27, 39), c_choose_fn(0.3, 27, 39))
        self.assertEqual(choose(0.7, 27, 39), c_choose_fn(0.7, 27, 39))

    def test_l2norm(self):
        mod = JitModule()
        submod = CFile("test_l2norm",
                       [l2norm_ast], path=CONFIG.get('jit','COMPILE_PATH'))._compile(l2norm_ast.codegen())
        mod._link_in(submod)
        entry = l2norm_ast.find(FunctionDecl, name="l2norm")
        c_l2norm_fn = mod.get_callable(entry.name, entry.get_type())
        self.assertEqual(l2norm(np.ones(12, dtype=np.float64)),
                         c_l2norm_fn(np.ones(12, dtype=np.float64), 12))

    def test_getFile(self):
        getFile(os.path.join(CONFIG.get('jit','COMPILE_PATH'),'test_l2norm.c'))

class TestAuxiliary(unittest.TestCase):
    def test_NameExtractor(self):
        def f(x):
            return x + 3

        py_ast = get_ast(f)
        result = LazySpecializedFunction.NameExtractor().visit(py_ast)
        self.assertEqual(result, 'f')

    def test_from_function(self):
        def f(x):
            return x + 3

        c_f = TestTranslator.from_function(f, 'test_from_function')
        self.assertEqual(c_f(3), 6)
